<?php

declare(strict_types=1);

namespace Drupal\commerce_paytrail\RequestBuilder;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Entity\OrderItemInterface;
use Drupal\commerce_paytrail\Event\ModelEvent;
use Drupal\commerce_paytrail\PaymentGatewayPluginTrait;
use Drupal\commerce_price\Calculator;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\commerce_price\Price;
use Drupal\commerce_product\Entity\ProductVariationInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Uuid\UuidInterface;
use Paytrail\SDK\Interfaces\ItemInterface;
use Paytrail\SDK\Model\CallbackUrl;
use Paytrail\SDK\Model\Customer;
use Paytrail\SDK\Model\Item;
use Paytrail\SDK\Request\AbstractPaymentRequest;
use Symfony\Component\DependencyInjection\Attribute\Autowire;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * A base class for payment requests.
 */
abstract class PaymentRequestBase {

  use PaymentGatewayPluginTrait;

  /**
   * Constructs a new instance.
   *
   * @param \Drupal\Component\Uuid\UuidInterface $uuidService
   *   The uuid service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   The event dispatcher.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $converter
   *   The minor unit converter.
   * @param int $callbackDelay
   *   The callback delay.
   */
  public function __construct(
    protected UuidInterface $uuidService,
    protected TimeInterface $time,
    protected EventDispatcherInterface $eventDispatcher,
    #[Autowire(value: '@commerce_price.minor_units_converter')] protected MinorUnitsConverterInterface $converter,
    #[Autowire(value: '%commerce_paytrail.callback_delay%')] protected int $callbackDelay,
  ) {
  }

  /**
   * Creates a Paytrail order item for given commerce order item.
   *
   * @param \Drupal\commerce_order\Entity\OrderItemInterface $orderItem
   *   The commerce order item.
   *
   * @return \Paytrail\SDK\Interfaces\ItemInterface
   *   The paytrail order item.
   */
  protected function createOrderLine(OrderItemInterface $orderItem) : ItemInterface {
    assert($orderItem->getAdjustedUnitPrice() instanceof Price);

    $item = (new Item())
      ->setUnitPrice($this->converter->toMinorUnits($orderItem->getAdjustedUnitPrice()))
      ->setUnits((int) $orderItem->getQuantity())
      ->setVatPercentage(0);

    if ($purchasedEntityId = $orderItem->getPurchasedEntityId()) {
      $item->setProductCode((string) $purchasedEntityId);
    }

    if ($taxes = $orderItem->getAdjustments(['tax'])) {
      $item->setVatPercentage((float) Calculator::multiply(
        (string) reset($taxes)->getPercentage(),
        '100'
      ));
    }

    // Override default product code with SKU if possible.
    if ($orderItem->getPurchasedEntity() instanceof ProductVariationInterface) {
      $item->setProductCode($orderItem->getPurchasedEntity()->getSku());
    }

    return $item;
  }

  /**
   * Populates the given payment request.
   *
   * @param \Paytrail\SDK\Request\AbstractPaymentRequest $request
   *   The base request.
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param string $event
   *   The event dispatcher event.
   *
   * @return \Paytrail\SDK\Request\AbstractPaymentRequest
   *   The payment request
   *
   * @throws \Drupal\commerce_paytrail\Exception\PaytrailPluginException
   */
  protected function populatePaymentRequest(AbstractPaymentRequest $request, OrderInterface $order, string $event) : AbstractPaymentRequest {
    $plugin = $this->getPaymentPlugin($order);

    $items = array_map(
      fn (OrderItemInterface $item) => $this->createOrderLine($item),
      $order->getItems()
    );
    assert($order->getTotalPrice() instanceof Price);

    $request->setAmount($this->converter->toMinorUnits($order->getTotalPrice()))
      ->setStamp($this->uuidService->generate())
      ->setLanguage($plugin->getLanguage())
      // Only EUR is supported.
      ->setCurrency('EUR')
      ->setCallbackUrls((new CallbackUrl())
        ->setSuccess($plugin->getNotifyUrl()->toString())
        ->setCancel($plugin->getNotifyUrl()->toString())
      )
      // Delay callback polling to minimize the chance of Paytrail
      // calling ::onNotify() before customer returns from the payment
      // gateway.
      ->setCallbackDelay($this->callbackDelay)
      ->setRedirectUrls((new CallbackUrl())
        ->setSuccess($plugin->getReturnUrl($order)->toString())
        ->setCancel($plugin->getCancelUrl($order)->toString())
      );

    $customer = (new Customer())
      ->setEmail($order->getEmail());

    $request->setCustomer($customer);

    // Add order level discounts as negative line item.
    foreach ($order->getAdjustments() as $adjustment) {
      if (!$adjustment->getAmount()->isNegative()) {
        continue;
      }
      $items[] = (new Item())
        ->setUnits(1)
        ->setVatPercentage(0)
        ->setProductCode('discount')
        ->setUnitPrice($this->converter->toMinorUnits($adjustment->getAmount()));
    }
    $request->setItems($items);

    $this
      ->eventDispatcher
      ->dispatch(new ModelEvent($request, $order, $event));
    // We use the reference field to load the order. Make sure it cannot be
    // changed.
    $request->setReference((string) $order->id());

    return $request;
  }

}
