<?php

declare(strict_types=1);

namespace Drupal\commerce_paytrail\RequestBuilder;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_paytrail\Event\ModelEvent;
use Drupal\commerce_paytrail\PaymentGatewayPluginTrait;
use Drupal\commerce_price\MinorUnitsConverterInterface;
use Drupal\commerce_price\Price;
use Drupal\Component\Uuid\UuidInterface;
use Paytrail\SDK\Model\CallbackUrl;
use Paytrail\SDK\Request\RefundRequest;
use Paytrail\SDK\Response\RefundResponse;
use Symfony\Component\DependencyInjection\Attribute\Autowire;
use Symfony\Contracts\EventDispatcher\EventDispatcherInterface;

/**
 * The refund request builder.
 *
 * @internal
 */
class RefundRequestBuilder implements RefundRequestBuilderInterface {

  use PaymentGatewayPluginTrait;

  /**
   * Constructs a new instance.
   *
   * @param \Drupal\Component\Uuid\UuidInterface $uuidService
   *   The uuid service.
   * @param \Symfony\Contracts\EventDispatcher\EventDispatcherInterface $eventDispatcher
   *   The event dispatcher.
   * @param \Drupal\commerce_price\MinorUnitsConverterInterface $converter
   *   The minor unit converter.
   */
  public function __construct(
    private UuidInterface $uuidService,
    private EventDispatcherInterface $eventDispatcher,
    #[Autowire(value: '@commerce_price.minor_units_converter')] protected MinorUnitsConverterInterface $converter,
  ) {
  }

  /**
   * Constructs a refund request.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   The order.
   * @param \Drupal\commerce_price\Price $amount
   *   The price.
   *
   * @return \Paytrail\SDK\Request\RefundRequest
   *   The refund request.
   */
  public function createRefundRequest(OrderInterface $order, Price $amount) : RefundRequest {
    $plugin = $this->getPaymentPlugin($order);

    $urls = new CallbackUrl();
    $urls->setCancel($plugin->getNotifyUrl(['event' => 'refund-cancel'])->toString())
      ->setSuccess($plugin->getNotifyUrl(['event' => 'refund-success'])->toString());

    return (new RefundRequest())
      ->setRefundReference((string) $order->id())
      ->setAmount($this->converter->toMinorUnits($amount))
      ->setCallbackUrls($urls)
      ->setRefundStamp($this->uuidService->generate());
  }

  /**
   * {@inheritdoc}
   */
  public function refund(string $transactionId, OrderInterface $order, Price $amount) : RefundResponse {
    $plugin = $this->getPaymentPlugin($order);

    $request = $this->createRefundRequest($order, $amount);
    $this->eventDispatcher
      ->dispatch(new ModelEvent(
        $request,
        $order,
        RefundRequestBuilderInterface::REFUND_CREATE
      ));

    $response = $plugin->getClient()
      ->refund($request, $transactionId);

    $this->eventDispatcher
      ->dispatch(new ModelEvent(
        $response,
        $order,
        RefundRequestBuilderInterface::REFUND_CREATE_RESPONSE
      ));

    return $response;
  }

}
