<?php

declare(strict_types=1);

namespace Drupal\Tests\commerce_paytrail\Unit;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentGatewayInterface as PaymentGatewayEntityInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\PaymentGatewayInterface;
use Drupal\commerce_paytrail\Exception\PaytrailPluginException;
use Drupal\commerce_paytrail\PaymentGatewayPluginTrait;
use Drupal\commerce_paytrail\Plugin\Commerce\PaymentGateway\PaytrailInterface;
use Drupal\Core\Entity\Plugin\DataType\EntityAdapter;
use Drupal\Core\Entity\Plugin\DataType\EntityReference;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\TypedData\ListInterface;
use Drupal\Tests\UnitTestCase;
use Prophecy\PhpUnit\ProphecyTrait;
use Prophecy\Prophecy\ObjectProphecy;

/**
 * Tests payment gateway plugin trait.
 */
class PaymentGatewayPluginTraitTest extends UnitTestCase {

  use PaymentGatewayPluginTrait;
  use ProphecyTrait;

  /**
   * Gets a mock for order object.
   *
   * @param \Prophecy\Prophecy\ObjectProphecy $gateway
   *   The gateway prophecy.
   *
   * @return \Prophecy\Prophecy\ObjectProphecy
   *   The order prophecy.
   */
  private function getMockOrder(ObjectProphecy $gateway) : ObjectProphecy {
    $entityAdapter = $this->prophesize(EntityAdapter::class);
    $entityAdapter->getValue()->willReturn($gateway->reveal());

    $entity = $this->prophesize(EntityReference::class);
    $entity->getTarget()->willReturn($entityAdapter->reveal());

    $entityReferenceList = $this->prophesize(EntityReferenceFieldItemListInterface::class);
    $entityReferenceList->get('entity')->willReturn($entity->reveal());

    $gatewayList = $this->prophesize(ListInterface::class);
    $gatewayList->isEmpty()->willReturn(FALSE);
    $gatewayList->first()->willReturn($entityReferenceList);

    $order = $this->prophesize(OrderInterface::class);
    $order->id()->willReturn(1);
    $order->get('payment_gateway')
      ->willReturn(
        $gatewayList->reveal()
      );
    return $order;
  }

  /**
   * Tests without payment gateway set.
   */
  public function testGatewayNotFoundException() : void {
    $list = $this->prophesize(ListInterface::class);
    $list->isEmpty()->willReturn(TRUE);

    $order = $this->prophesize(OrderInterface::class);
    $order->id()->willReturn(1);
    $order->get('payment_gateway')
      ->willReturn(
        $list->reveal()
      );
    $this->expectException(PaytrailPluginException::class);
    $this->expectExceptionMessage('Payment gateway not found.');
    $this->getPaymentPlugin($order->reveal());
  }

  /**
   * Tests with invalid plugin.
   */
  public function testInvalidPluginException() : void {
    $gateway = $this->prophesize(PaymentGatewayEntityInterface::class);
    $gateway->getPlugin()->willReturn($this->prophesize(PaymentGatewayInterface::class)->reveal());

    $order = $this->getMockOrder($gateway);

    $this->expectException(PaytrailPluginException::class);
    $this->expectExceptionMessage('Payment gateway not instanceof PaytrailInterface.');
    $this->getPaymentPlugin($order->reveal());
  }

  /**
   * Tests with valid payment plugin.
   */
  public function testGetPaymentPlugin() : void {
    $plugin = $this->prophesize(PaytrailInterface::class);
    $gateway = $this->prophesize(PaymentGatewayEntityInterface::class);
    $gateway->getPlugin()->willReturn($plugin->reveal());
    $order = $this->getMockOrder($gateway);

    static::assertInstanceOf(PaytrailInterface::class, $this->getPaymentPlugin($order->reveal()));
  }

}
